import netCDF4
from paegan.cdm.dataset import CommonDataset
from paegan.transport.location4d import Location4D

class Bathymetry():
    def __init__(self, **kwargs):
        """
            Optional named arguments: 
            * file (local path or dap to bathymetry netcdf file)

        """
        
        if "file" in kwargs:
            self._file = os.path.normpath(kwargs.pop('file'))
        else:
            self._file = os.path.normpath(os.path.join(__file__,"../../resources/bathymetry/ETOPO1_Bed_g_gmt4.grd"))
        
        self._type = kwargs.pop("type", "hover")
        self._nc = CommonDataset(self._file)
        self._bathy_name = kwargs.pop("bathy", "z")

    def intersect(self, **kwargs):
        """
            Intersect Point and Bathymetry
            returns bool
        """
        end_point = kwargs.pop('end_point')
        depth = np.mean(np.mean(self._nc.get_values(self._bathy_name, point=end_point)))
        if depth > end_point.depth:
            inter = True
        else:
            inter = False
        return inter
        
    def react(self, **kwargs):
    
        if self._type == 'hover':
            return self.__hover(**kwargs)
        elif self._type == 'stick':
            pass
        else:
            raise ValueError("Bathymetry interaction type not supported")
            
    def __hover(self, **kwargs):
        end_point = kwargs.pop('end_point')
        depth = np.mean(np.mean(self._nc.get_values(self._bathy_name, point=end_point)))
        return Location4D(latitude=end_point.latitude, longitude=end_point.longitude, depth=depth + 1)
        
        



from shapely.geometry import LineString

class Particle(object):
    """
        A particle
    """
    def __init__(self, **kwargs):
        self._locations = []
        self._age = 0. # Age in days
        self._id = kwargs.get("id", -1)

    def get_id(self):
        return self._id
    uid = property(get_id)

    def set_location(self, location):
        self._locations.append(location)
    def get_location(self):
        return self._locations[-1]
    location = property(get_location, set_location)

    def get_locations(self):
        return self._locations
    locations = property(get_locations, None)

    def fill_location_gap(self):
        """
        To keep certain parameters the same length as 'locations',
        this is used if the particle moves twice within a timestep.
        This can happen if they hit a shoreline, etc.
        """
        pass

    def set_active(self, active):
        self._active = active
    def get_active(self):
        return self._active
    active = property(get_active, set_active)

    def get_last_movement(self):
        return LineString(list(self.locations[-2].point.coords) + list(self.locations[-1].point.coords))

    def get_age(self, **kwargs):
        """
        Returns the particlees age (how long it has been forced) in a variety of units.
        Rounded to 8 decimal places.

        Parameters:
            units (optional) = 'days' (default), 'hours', 'minutes', or 'seconds'
        """
        try:
            units = kwargs.get('units', None)
            if units is None:
                return self._age
            units = units.lower()
            if units == "days":
                z = self._age
            elif units == "hours":
                z = self._age * 24
            elif units == "minutes":
                z = self._age * 24 * 60
            elif units == "seconds":
                z = self._age * 24 * 60 * 60
            else:
                raise
            return round(z,8) 
        except:
            raise KeyError("Could not return age of particle")

    def age(self, **kwargs):
        """
        Age this particle.

        parameters (optional, only one allowed):
            days (default)
            hours
            minutes
            seconds
        """
        if kwargs.get('days', None) is not None:
            self._age += kwargs.get('days')
            return
        if kwargs.get('hours', None) is not None:
            self._age += kwargs.get('hours') / 24.
            return
        if kwargs.get('minutes', None) is not None:
            self._age += kwargs.get('minutes') / 24. / 60.
            return
        if kwargs.get('seconds', None) is not None:
            self._age += kwargs.get('seconds') / 24. / 60. / 60.
            return

        raise KeyError("Could not age particle, please specify 'days', 'hours', 'minutes', or 'seconds' parameter")

    def linestring(self):
        return LineString(map(lambda x: list(x.point.coords)[0], self.locations))
        
class LarvaParticle(Particle):
    """
        A particle for larvamap, keeps track of information
        for the behaviors component
    """
    def __init__(self, **kwargs):
        super(LarvaParticle,self).__init__(**kwargs)
        self.lifestage_progress = 0.
        self._temp = []
        self._salt = []

    def set_temp(self, temp):
        self._temp.append(temp)
    def get_temp(self):
        return self._temp[-1]
    temp = property(get_temp, set_temp)
    
    def get_temps(self):
        return self._temp
    temps = property(get_temps, None)
    
    def set_salt(self, salt):
        self._salt.append(salt)
    def get_salt(self):
        return self._salt[-1]
    salt = property(get_salt, set_salt)
    
    def get_salts(self):
        return self._salt
    salts = property(get_salts, None)
    
    def get_lifestage_index(self):
        return int(self.lifestage_progress)
    lifestage_index = property(get_lifestage_index, None)

    def fill_location_gap(self):
        self._temp.append(self.temp)
        self._salt.append(self.salt)

    def grow(self, amount):
        """
        Grow a particle by a percentage value (0 < x < 1)

        When a particle grows past 1, its current lifestage is 
        complete and it moves onto the next.

        The calculation to get the current lifestage index is in get_lifestage_index()
        """
        self.lifestage_progress += amount
        
class ChemistryParticle(Particle):
    """
        A chemical particle for time and chemistry dependent
        tracers
    """
    def __init__(self):
        pass
        

class OilParticle(Particle):
    """
        An oil particle for special oil physics and chemistry
    """
    def __init__(self):
        pass



import unittest
import time
import numpy as np
import netCDF4
from datetime import datetime, timedelta
from paegan.transport.models.transport import Transport
from paegan.transport.particles.particle import Particle
from paegan.transport.location4d import Location4D
from paegan.utils.asarandom import AsaRandom
from paegan.transport.shoreline import Shoreline
from paegan.transport.bathymetry import Bathymetry
from multiprocessing import Value
import multiprocessing
from paegan.cdm.dataset import CommonDataset
import os
import random
#from paegan.logger import queue_logger

class Consumer(multiprocessing.Process):
    def __init__(self, task_queue, result_queue, n_run, lock):
        """
            This is the process class that does all the handling of queued tasks
        """
        multiprocessing.Process.__init__(self)
        self.task_queue = task_queue
        self.result_queue = result_queue
        self.n_run = n_run
        self.lock = lock
        lock.acquire()
        self.n_run.value = self.n_run.value + 1
        lock.release()
        
    def run(self):
        proc_name = self.name
        while True:
            next_task = self.task_queue.get()
            #queue_logger.start()
            if next_task is None:
                # Poison pill means shutdown
                #print '%s: Exiting' % proc_name
                #queue_logger.logger().info("Particle " + proc_name + " at exit")
                #queue_logger.logger().info(str(self.n_run.value) + " remaining processes bfore")
                self.lock.acquire()
                self.n_run.value = self.n_run.value - 1
                #queue_logger.logger().info(str(self.n_run.value) + " remaining processes after")
                self.lock.release()
                self.task_queue.task_done()
                #queue_logger.stop()
                
                break
            #print '%s: %s' % (proc_name, next_task)
            answer = next_task(proc_name)
            self.task_queue.task_done()
            self.result_queue.put(answer)
        return

class DataController(object):
    def __init__(self, url, n_run, get_data, updating,
                 init_size, horiz_chunk, particle_get, times,
                 start_time, point_get, start,
                 **kwargs
                 ):
        """
            The data controller controls the updating of the
            local netcdf data cache
        """
        assert "cache" in kwargs
        self.cache_path = kwargs["cache"]
        self.url = url
        #self.local = Dataset(".cache/localcache.nc", 'w')
        self.n_run = n_run
        self.get_data = get_data
        self.updating = updating
        self.inds = None#np.arange(init_size+1)
        self.time_size = init_size
        self.horiz_size = horiz_chunk
        self.particle_get = particle_get
        self.point_get = point_get
        self.low_memory = kwargs.get("low_memory", False)
        self.start_time = start_time
        self.times = times
        
        self.start = start

    
    def get_variablenames_for_model(self):
        getname = self.dataset.get_varname_from_stdname
        self.uname = getname('eastward_sea_water_velocity') 
        self.vname = getname('northward_sea_water_velocity') 
        self.wname = getname('upward_sea_water_velocity')
        if len(self.uname) > 0:
            self.uname = self.uname[0]
        else:
            self.uname = None
        if len(self.vname) > 0:
            self.vname = self.vname[0]
        else:
            self.vname = None
        if len(self.wname) > 0:
            self.wname = self.wname[0]
        else:
            self.wname = None

        coords = self.dataset.get_coord_names(self.uname) 
        self.xname = coords['xname'] 
        self.yname = coords['yname']
        self.zname = coords['zname']
        self.tname = coords['tname']
        self.temp_name = getname('sea_water_temperature') 
        self.salt_name = getname('sea_water_salinity')
        
        if len(self.temp_name) > 0:
            self.temp_name = self.temp_name[0]
        else:
            self.temp_name = None       
        if len(self.salt_name) > 0:
            self.salt_name = self.salt_name[0]
        else:
            self.salt_name = None
        self.tname = None ## temporary
    
    def get_remote_data(self, localvars, remotevars, inds, shape):
        """
            Method that does the updating of local netcdf cache
            with remote data
        """
        # If user specifies 'all' then entire xy domain is
        # grabbed, default is 4, specified in the model_controller
        if self.horiz_size == 'all':
            y, y_1 = 0, shape[-2]
            x, x_1 = 0, shape[-1]
        else:
            r = self.horiz_size
            x, x_1 = self.point_get.value[2]-r, self.point_get.value[2]+r+1
            y, y_1 = self.point_get.value[1]-r, self.point_get.value[1]+r+1
            x, x_1 = x[0], x_1[0]
            y, y_1 = y[0], y_1[0]
            if y < 0:
                y = 0
            if x < 0:
                x = 0
            if y_1 > shape[-2]:
                y_1 = shape[-2]
            if x_1 > shape[-1]:
                x_1 = shape[-1]
        
        # Update domain variable for where we will add data
        domain = self.local.variables['domain']
        if self.low_memory:
            for z in range(shape[1]):
                if z + 1 > shape[1] - 1:
                    z_1 = shape[1]
                else:
                    z_1 = z + 1
                if len(shape) == 4:
                    domain[time:time_1, z:z_1, y:y_1, x:x_1] = np.ones((time_1-time, z_1-z, y_1-y, x_1-x))
                elif len(shape) == 3:
                    if z == self.inds[0]:
                        domain[time:time_1, y:y_1, x:x_1] = np.ones((time_1-time, y_1-y, x_1-x))
        else:
            if len(shape) == 4:
                domain[inds[0]:inds[-1]+1, 0:shape[1], y:y_1, x:x_1] = np.ones((inds[-1]+1-inds[0], shape[1], y_1-y, x_1-x))
            elif len(shape) == 3:
                domain[inds[0]:inds[-1]+1, y:y_1, x:x_1] = np.ones((inds[-1]+1-inds[0], y_1-y, x_1-x))
        
        # Update the local variables with remove data              
        for local, remote in zip(localvars, remotevars):
            if self.low_memory:
                for z in range(shape[1]):
                    if z + 1 > shape[1] - 1:
                        z_1 = shape[1]
                    else:
                        z_1 = z + 1
                    if len(shape) == 4:
                        local[time:time_1, z:z_1, y:y_1, x:x_1] = remote[time:time_1,  z:z_1, y:y_1, x:x_1]
                    else:
                        if z == 0:
                            local[time:time_1, y:y_1, x:x_1] = remote[time:time_1, y:y_1, x:x_1]
           
            else:
                #print "time", inds, "rem", shape[1], y, y_1, x, x_1
                if len(shape) == 4:
                    local[inds[0]:inds[-1]+1, 0:shape[1], y:y_1, x:x_1] = remote[inds[0]:inds[-1]+1,  0:shape[1], y:y_1, x:x_1]
                else:
                    local[inds[0]:inds[-1]+1, y:y_1, x:x_1] = remote[inds[0]:inds[-1]+1, y:y_1, x:x_1]

    def __call__(self, proc):
        c = 0
        self.dataset = CommonDataset(self.url)
        self.proc = proc
        self.get_variablenames_for_model()
        self.remote = self.dataset.nc
        cachepath = self.cache_path
        times = self.times
        start_time = self.start_time
        
        # Calculate the datetimes of the model timesteps like
        # the particle objects do, so we can figure out unique
        # time indices
        modelTimestep = []
        calculatedTime = []
        newtimes = []
        for i in xrange(0, len(times)-1):
            try:
                modelTimestep.append(times[i+1] - times[i])
                calculatedTime.append(times[i+1])
            except:
                modelTimestep.append(times[i] - times[i-1])
                calculatedTime.append(times[i] + modelTimestep[i])
               
            newtimes.append(start_time + timedelta(seconds=calculatedTime[i]))
      
        timevar = self.dataset.gettimevar(self.uname)
        time_indexs = timevar.nearest_index(newtimes)
        
        self.inds = np.unique(time_indexs)
        
        # While there is at least 1 particle still running, 
        # stay alive, if not break
        while self.n_run.value > 1:
            # If particle asks for data, do the following
            if self.get_data.value == True:
                if c == 0:
                    indices = self.dataset.get_indices(self.uname, timeinds=[np.asarray([0])], point=self.start)
                    self.point_get.value = [self.inds[0], indices[-2], indices[-1]]
                    
                    # Set shared state so particles know the 
                    # cache is being updated
                    self.updating.value = True
                    
                    # Wait for particles to get out
                    while self.particle_get == True:
                        pass
                    
                    # Open local cache for writing, overwrites
                    # existing file with same name
                    self.local = netCDF4.Dataset(cachepath, 'w')
                    
                    # Create dimensions for u and v variables
                    self.local.createDimension('time', None)
                    self.local.createDimension('level', None)
                    self.local.createDimension('x', None)
                    self.local.createDimension('y', None)
                    
                    # Create 3d or 4d u and v variables
                    if self.remote.variables[self.uname].ndim == 4:
                        self.ndim = 4
                        dimensions = ('time', 'level', 'y', 'x')
                        coordinates = "time z lon lat"
                    elif self.remote.variables[self.uname].ndim == 3:
                        self.ndim = 3
                        dimensions = ('time', 'y', 'x')
                        coordinates = "time lon lat"
                    shape = self.remote.variables[self.uname].shape
                    try:
                        fill = self.remote.variables[self.uname].missing_value
                    except:
                        fill = None
                    
                    # Create domain variable that specifies
                    # where there is data geographically/by time
                    # and where there is not data,
                    #   Used for testing if particle needs to 
                    #   ask cache to update
                    domain = self.local.createVariable('domain',
                            'i', dimensions, zlib=False, fill_value=0,
                            )
                    domain.coordinates = coordinates
                            
                    if fill == None:
                        # Create local u and v variables
                        u = self.local.createVariable('u', 
                            'f', dimensions, zlib=False,
                            )
                        v = self.local.createVariable('v', 
                            'f', dimensions, zlib=False,
                            )
                        
                        v.coordinates = coordinates
                        u.coordinates = coordinates
                        
                        # Create local w variable
                        if self.wname != None:
                            w = self.local.createVariable('w', 
                                'f', dimensions, zlib=False,
                                )
                            w.coordinates = coordinates
                        if self.temp_name != None and self.salt_name != None:      
                            # Create local temp and salt vars  
                            temp = self.local.createVariable('temp', 
                                'f', dimensions, zlib=False,
                                )
                            salt = self.local.createVariable('salt', 
                                'f', dimensions, zlib=False,
                                )
                            temp.coordinates = coordinates
                            salt.coordinates = coordinates
                    else:    
                        # Create local u and v variables
                        u = self.local.createVariable('u', 
                            'f', dimensions, zlib=False,
                            fill_value=fill)
                        v = self.local.createVariable('v', 
                            'f', dimensions, zlib=False,
                            fill_value=fill)
                        
                        v.coordinates = coordinates
                        u.coordinates = coordinates
                        
                        # Create local w variable
                        if self.wname != None:
                            w = self.local.createVariable('w', 
                                'f', dimensions, zlib=False,
                                fill_value=fill)
                            w.coordinates = coordinates
                        if self.temp_name != None and self.salt_name != None: 
                            # Create local temp and salt vars       
                            temp = self.local.createVariable('temp', 
                                'f', dimensions, zlib=False,
                                fill_value=fill)
                            salt = self.local.createVariable('salt', 
                                'f', dimensions, zlib=False,
                                fill_value=fill)
                            temp.coordinates = coordinates
                            salt.coordinates = coordinates
                    
                    # Create local lat/lon coordinate variables
                    if self.remote.variables[self.xname].ndim == 2:
                        lon = self.local.createVariable('lon',
                                'f', ("y", "x"), zlib=False,
                                )
                        lat = self.local.createVariable('lat',
                                'f', ("y", "x"), zlib=False,
                                )
                    if self.remote.variables[self.xname].ndim == 1:
                        lon = self.local.createVariable('lon',
                                'f', ("x"), zlib=False,
                                )
                        lat = self.local.createVariable('lat',
                                'f', ("y"), zlib=False,
                                )
                    
                    if self.remote.variables[self.xname].ndim == 2:             
                        lon[:] = self.remote.variables[self.xname][:, :]
                        lat[:] = self.remote.variables[self.yname][:, :]
                    if self.remote.variables[self.xname].ndim == 1:
                        lon[:] = self.remote.variables[self.xname][:]
                        lat[:] = self.remote.variables[self.yname][:]
                    
                    localvars = [u, v,]
                    remotevars = [self.remote.variables[self.uname], 
                                  self.remote.variables[self.vname]]
                                  
                    if self.temp_name != None and self.salt_name != None:
                        localvars.append(temp)
                        localvars.append(salt)
                        remotevars.append(self.remote.variables[self.temp_name])
                        remotevars.append(self.remote.variables[self.salt_name])
                    if self.wname != None:
                        localvars.append(w)
                        remotevars.append(self.remote.variables[self.wname])
                        
                    # Create local z variable
                    if self.zname != None:            
                        if self.remote.variables[self.zname].ndim == 4:
                            z = self.local.createVariable('z',
                                'f', ("time","level","y","x"), zlib=False,
                                )  
                            remotez = self.remote.variables[self.zname]
                            localvars.append(z)
                            remotevars.append(remotez)
                        elif self.remote.variables[self.zname].ndim == 3:
                            z = self.local.createVariable('z',
                                'f', ("level","y","x"), zlib=False,
                                )
                            z[:] = self.remote.variables[self.zname][:, :, :]
                        elif self.remote.variables[self.zname].ndim ==1:
                            z = self.local.createVariable('z',
                                'f', ("level",), zlib=False,
                                )
                            z[:] = self.remote.variables[self.zname][:]
                            
                    # Create local time variable
                    time = self.local.createVariable('time',
                                'f8', ("time",), zlib=False,
                                )
                    if self.tname != None:
                        time[:] = self.remote.variables[self.tname][self.inds]
                    
                    if self.point_get.value[0]+self.time_size > np.max(self.inds):
                        current_inds = np.arange(self.point_get.value[0], np.max(self.inds)+1)
                    else:
                        current_inds = np.arange(self.point_get.value[0],self.point_get.value[0] + self.time_size)
                    
                    # Get data from remote dataset and add
                    # to local cache  
                    self.get_remote_data(localvars, remotevars, current_inds, shape) 
                    
                    self.local.sync()
                    self.local.close()
                    c += 1

                    self.updating.value = False
                    self.get_data.value = False

                else:
                    #print "updating appending"
                    self.updating.value = True
                    # Wait for particles to get out (this is
                    # poorly implemented)
                    while self.particle_get == True:
                        pass
                        
                    # Open local cache dataset for appending
                    self.local = netCDF4.Dataset(cachepath, 'a')
                    
                    # Create local and remote variable objects
                    # for the variables of interest  
                    u = self.local.variables['u']
                    v = self.local.variables['v']
                    time = self.local.variables['time']
                    remoteu = self.remote.variables[self.uname]
                    remotev = self.remote.variables[self.vname]
                    
                    # Create lists of variable objects for
                    # the data updater
                    localvars = [u, v, ]
                    remotevars = [remoteu, remotev, ]
                    if self.salt_name != None and self.temp_name != None:
                        salt = self.local.variables['salt']
                        temp = self.local.variables['temp']
                        remotesalt = self.remote.variables[self.salt_name]
                        remotetemp = self.remote.variables[self.temp_name]
                        localvars.append(salt)
                        localvars.append(temp)
                        remotevars.append(remotesalt)
                        remotevars.append(remotetemp)
                    if self.wname != None:
                        w = self.local.variables['w']
                        remotew = self.remote.variables[self.wname]
                        localvars.append(w)
                        remotevars.append(remotew)
                    if self.zname != None:
                        remotez = self.remote.variables[self.zname]
                        if remotez.ndim == 4:
                            z = self.local.variables['z']
                            localvars.append(z)
                            remotevars.append(remotez)
                    if self.tname != None:
                        remotetime = self.remote.variables[self.tname]
                        time[self.inds] = self.remote.variables[self.inds]
                    
                    if self.point_get.value[0]+self.time_size > np.max(self.inds):
                        current_inds = np.arange(self.point_get.value[0], np.max(self.inds)+1)
                    else:
                        current_inds = np.arange(self.point_get.value[0],self.point_get.value[0] + self.time_size)
                    
                    # Get data from remote dataset and add
                    # to local cache
                    self.get_remote_data(localvars, remotevars, current_inds, shape)
                    
                    self.local.sync()
                    self.local.close()
                    c += 1

                    self.updating.value = False
                    self.get_data.value = False
            else:
                pass
        return
            
        
class ForceParticle(object):
    from paegan.transport.shoreline import Shoreline
    from paegan.transport.bathymetry import Bathymetry
    #from paegan.cdm.dataset import CommonDataset
    def __str__(self):
        return self.part.__str__()

    def __init__(self, part, remotehydro, times, start_time, models, 
                 point, usebathy, useshore, usesurface,
                 get_data, n_run, updating, particle_get,
                 point_get, request_lock, cache=None):
        """
            This is the task/class/object/job that forces an
            individual particle and communicates with the 
            other particles and data controller for local
            cache updates
        """
        assert cache != None
        self.cache_path = cache
        self.remotehydropath = remotehydro
        self.localpath =  self.cache_path
        self.point = point
        self.part = part
        self.times = times
        self.start_time = start_time
        self.models = models
        self.usebathy = usebathy
        self.useshore = useshore
        self.usesurface = usesurface
        self.get_data = get_data
        self.n_run = n_run
        self.updating = updating
        self.particle_get = particle_get
        self.point_get = point_get
        self.request_lock = request_lock
        
        
    def get_variablenames_for_model(self, dataset):
        # Use standard names to get variable names for required
        # model parameters
        getname = dataset.get_varname_from_stdname
        self.uname = getname('eastward_sea_water_velocity') 
        self.vname = getname('northward_sea_water_velocity') 
        self.wname = getname('upward_sea_water_velocity')
        if len(self.uname) > 0:
            self.uname = self.uname[0]
        else:
            self.uname = None
        if len(self.vname) > 0:
            self.vname = self.vname[0]
        else:
            self.vname = None
        if len(self.wname) > 0:
            self.wname = self.wname[0]
        else:
            self.wname = None
        
        # Get coordinate names based on u variable
        coords = dataset.get_coord_names(self.uname) 
        self.xname = coords['xname'] 
        self.yname = coords['yname']
        self.zname = coords['zname']
        self.tname = coords['tname']
        # Get salt and temp names from std names
        self.temp_name = getname('sea_water_temperature') 
        self.salt_name = getname('sea_water_salinity')
        
        if len(self.temp_name) > 0:
            self.temp_name = self.temp_name[0]
        else:
            self.temp_name = None       
        if len(self.salt_name) > 0:
            self.salt_name = self.salt_name[0]
        else:
            self.salt_name = None
        self.tname = None ## temporary
        
    def need_data(self, i):
        """
            Method to test if cache contains the data that
            the particle needs
        """
        self.particle_get.value = True
        self.dataset.opennc()
        # Test if the cache has the data we need
        try:
            # If the point we request contains fill values, 
            # we need data
            if type(np.mean(np.mean(self.dataset.get_values('domain', timeinds=[np.asarray([i])], point=self.part.location )))) == np.ma.core.MaskedArray:
                need = True
            else:
                need = False
        except:
            # If the time index doesnt even exist, we need
            need = True
        self.dataset.closenc()
        self.particle_get.value = False
        #print self.proc, need
        return need # return true if need data or false if dont
      
    def data(self, i):
        """
            Method to streamline request for data from cache
        """
        while self.get_data.value == True:
            pass
        #print self.proc, "done waiting"
        if self.need_data(i):
            #print self.proc, "yes i do need data"
            # Acquire lock for asking for data
            self.request_lock.acquire()
            if self.need_data(i):
                # Open netcdf file on disk from commondataset
                self.dataset.opennc()
                # Get the indices for the current particle location
                indices = self.dataset.get_indices('u', timeinds=[np.asarray([i-1])], point=self.part.location )
                self.dataset.closenc()
                # Override the time
                self.point_get.value = [indices[0]+1, indices[-2], indices[-1]]
                # Request that the data controller update the cache
                self.get_data.value = True
                # Wait until the data controller is done
                while self.get_data.value == True:
                    pass 
            self.request_lock.release()
               
        # Announce that someone is getting data from the local
        self.particle_get.value = True
        # Open netcdf file on disk from commondataset
        self.dataset.opennc()
        # Grab data at time index closest to particle location
        u = np.mean(np.mean(self.dataset.get_values('u', timeinds=[np.asarray([i])], point=self.part.location )))
        v = np.mean(np.mean(self.dataset.get_values('v', timeinds=[np.asarray([i])], point=self.part.location )))
        # if there is vertical velocity inthe dataset, get it
        if 'w' in self.dataset.nc.variables:
            w = np.mean(np.mean(self.dataset.get_values('w', timeinds=[np.asarray([i])], point=self.part.location )))
        else:
            w = 0.0
        # If there is salt and temp in the dataset, get it
        if self.temp_name != None and self.salt_name != None:
            temp = np.mean(np.mean(self.dataset.get_values('temp', timeinds=[np.asarray([i])], point=self.part.location )))
            salt = np.mean(np.mean(self.dataset.get_values('salt', timeinds=[np.asarray([i])], point=self.part.location )))
        
        # Check for nans that occur in the ocean (happens because
        # of model and coastline resolution mismatches)
        if np.isnan(u) or np.isnan(v) or np.isnan(w):
            # Take the mean of the closest 4 points
            # If this includes nan which it will, result is nan
            u = np.mean(np.mean(self.dataset.get_values('u', timeinds=[np.asarray([i])], point=self.part.location, num=4)))
            v = np.mean(np.mean(self.dataset.get_values('v', timeinds=[np.asarray([i])], point=self.part.location, num=4)))
            if 'w' in self.dataset.nc.variables:
                w = np.mean(np.mean(self.dataset.get_values('w', timeinds=[np.asarray([i])], point=self.part.location, num=4)))
            else:
                w = 0.0
            if self.temp_name != None and self.salt_name != None:
                temp = np.mean(np.mean(self.dataset.get_values('temp', timeinds=[np.asarray([i])], point=self.part.location, num=4)))
                salt = np.mean(np.mean(self.dataset.get_values('salt', timeinds=[np.asarray([i])], point=self.part.location, num=4)))
        # If the final data results are nan, set velocities to zero
        # TODO: need to validate this implementation
        if np.isnan(u) or np.isnan(v) or np.isnan(w):
            u, v, w = 0.0, 0.0, 0.0
        self.dataset.closenc()
        self.particle_get.value = False
        if self.temp_name != None and self.salt_name != None:
            return u, v, w, temp, salt
        else:
            return u,v,w, None, None
        
    def __call__(self, proc):
        if self.usebathy == True:
            self._bathymetry = Bathymetry(point=self.point)
        else:
            self._bathymetry = None
        if self.useshore == True:
            self._shoreline = Shoreline(point=self.point)
        else:
            self._shoreline = None
        self.proc = proc
        part = self.part
        times = self.times
        start_time = self.start_time
        models = self.models
        
        while self.get_data.value == True:
            pass
        # Initialize commondataset of local cache, then
        # close the related netcdf file
        self.dataset = CommonDataset(self.localpath)
        self.dataset.closenc()
        
        # Calculate datetime at every timestep
        modelTimestep = []
        calculatedTime = []
        newtimes = []
        for i in xrange(0, len(times)-1):
            try:
                modelTimestep.append(times[i+1] - times[i])
                calculatedTime.append(times[i+1])
            except:
                modelTimestep.append(times[i] - times[i-1])
                calculatedTime.append(times[i] + modelTimestep[i])
               
            newtimes.append(start_time + timedelta(seconds=calculatedTime[i]))
        
        # Get variable names for parameters required by model
        remote = CommonDataset(self.remotehydropath)
        self.get_variablenames_for_model(remote)

        # Figure out indices corresponding to timesteps
        timevar = remote.gettimevar(self.uname)
        time_indexs = timevar.nearest_index(newtimes)
        array_indexs = time_indexs - time_indexs[0]

        # loop over timesteps   
        for loop_i, i in enumerate(time_indexs):
     
            #queue_logger.logger().info("Particle %i at %s" % (part.id, newtimes[loop_i].isoformat()))

            newloc = None
            
            
            # Get data from local netcdf here
            # dataset = CommonDataset(".cache/localcache.nc")
            
            # if need a time that is outside of what we have:e
            while self.get_data.value == True:
                pass
                
            # Get the variable data required by the models
            u, v, w, temp, salt = self.data(i)

            # Age the particle by the modelTimestep (seconds)
            # 'Age' meaning the amount of time it has been forced.
            part.age(seconds=modelTimestep[loop_i])

            # loop over models - sort these in the order you want them to run
            for model in models:
                movement = model.move(part, u, v, w, modelTimestep[loop_i], temperature=temp, salinity=salt)
                newloc = Location4D(latitude=movement['latitude'], longitude=movement['longitude'], depth=movement['depth'], time=newtimes[loop_i])
                if newloc: # changed p.location to part.location
                    self.boundary_interaction(self._bathymetry, self._shoreline, self.usebathy,self.useshore,self.usesurface,
                        particle=part, starting=part.location, ending=newloc,
                        distance=movement['distance'], angle=movement['angle'], 
                        azimuth=movement['azimuth'], reverse_azimuth=movement['reverse_azimuth'], 
                        vertical_distance=movement['vertical_distance'], vertical_angle=movement['vertical_angle'])
        return part
    
    def boundary_interaction(self, bathy, shore, usebathy, useshore, usesurface,
                             **kwargs):
        """
            Returns a list of Location4D objects
        """
        particle = kwargs.pop('particle')
        starting = kwargs.pop('starting')
        ending = kwargs.pop('ending')

        # shoreline
        if useshore:
            intersection_point = shore.intersect(start_point=starting.point, end_point=ending.point)
            if intersection_point:
                # Set the intersection point
                hitpoint = Location4D(point=intersection_point['point'], time=starting.time + (ending.time - starting.time))
                particle.location = hitpoint
                particle.fill_location_gap()
                resulting_point = shore.react(start_point=starting,
                                              end_point=ending,
                                              hit_point=hitpoint,
                                              feature=intersection_point['feature'],
                                              distance=kwargs.get('distance'),
                                              angle=kwargs.get('angle'),
                                              azimuth=kwargs.get('azimuth'),
                                              reverse_azimuth=kwargs.get('reverse_azimuth'))
                ending.latitude = resulting_point.latitude
                ending.longitude = resulting_point.longitude
                ending.depth = resulting_point.depth

        # bathymetry
        if usebathy:
            bintersect = bathy.intersect(start_point=starting,
                                         end_point=ending,
                                        )
            if bintersect:
                pt = bathy.react(type='hover', end_point=ending)
                ending.latitude = pt.latitude
                ending.longitude = pt.longitude
                ending.depth = pt.depth

        # sea-surface
        if usesurface:
            if ending.depth > 0:
                ending.depth = 0

        particle.location = ending
    



import logging
try:
    import Queue as queue
except ImportError:
    import queue
import threading

class QueueHandler(logging.Handler):
    """
    This handler sends events to a queue. Typically, it would be used together
    with a multiprocessing Queue to centralise logging to file in one process
    (in a multi-process application), so as to avoid file write contention
    between processes.

    This code is new in Python 3.2, but this class can be copy pasted into
    user code for use with earlier Python versions.
    """

    def __init__(self, queue):
        """
        Initialise an instance, using the passed queue.
        """
        logging.Handler.__init__(self)
        self.queue = queue

    def enqueue(self, record):
        """
        Enqueue a record.

        The base implementation uses put_nowait. You may want to override
        this method if you want to use blocking, timeouts or custom queue
        implementations.
        """
        self.queue.put_nowait(record)

    def prepare(self, record):
        """
        Prepares a record for queuing. The object returned by this method is
        enqueued.

        The base implementation formats the record to merge the message
        and arguments, and removes unpickleable items from the record
        in-place.

        You might want to override this method if you want to convert
        the record to a dict or JSON string, or send a modified copy
        of the record while leaving the original intact.
        """
        # The format operation gets traceback text into record.exc_text
        # (if there's exception data), and also puts the message into
        # record.message. We can then use this to replace the original
        # msg + args, as these might be unpickleable. We also zap the
        # exc_info attribute, as it's no longer needed and, if not None,
        # will typically not be pickleable.
        self.format(record)
        record.msg = record.message
        record.args = None
        record.exc_info = None
        return record

    def emit(self, record):
        """
        Emit a record.

        Writes the LogRecord to the queue, preparing it for pickling first.
        """
        try:
            self.enqueue(self.prepare(record))
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            self.handleError(record)

class QueueListener(object):
    """
    This class implements an internal threaded listener which watches for
    LogRecords being added to a queue, removes them and passes them to a
    list of handlers for processing.
    """
    _sentinel = None

    def __init__(self, queue, *handlers):
        """
        Initialise an instance with the specified queue and
        handlers.
        """
        self.queue = queue
        self.handlers = handlers
        self._stop = threading.Event()
        self._thread = None

    def dequeue(self, block):
        """
        Dequeue a record and return it, optionally blocking.

        The base implementation uses get. You may want to override this method
        if you want to use timeouts or work with custom queue implementations.
        """
        return self.queue.get(block)

    def start(self):
        """
        Start the listener.

        This starts up a background thread to monitor the queue for
        LogRecords to process.
        """
        self._thread = t = threading.Thread(target=self._monitor)
        t.setDaemon(True)
        t.start()

    def prepare(self , record):
        """
        Prepare a record for handling.

        This method just returns the passed-in record. You may want to
        override this method if you need to do any custom marshalling or
        manipulation of the record before passing it to the handlers.
        """
        return record

    def handle(self, record):
        """
        Handle a record.

        This just loops through the handlers offering them the record
        to handle.
        """
        record = self.prepare(record)
        for handler in self.handlers:
            handler.handle(record)

    def _monitor(self):
        """
        Monitor the queue for records, and ask the handler
        to deal with them.

        This method runs on a separate, internal thread.
        The thread will terminate if it sees a sentinel object in the queue.
        """
        q = self.queue
        has_task_done = hasattr(q, 'task_done')
        while not self._stop.isSet():
            try:
                record = self.dequeue(True)
                if record is self._sentinel:
                    break
                self.handle(record)
                if has_task_done:
                    q.task_done()
            except queue.Empty:
                pass
        # There might still be records in the queue.
        while True:
            try:
                record = self.dequeue(False)
                if record is self._sentinel:
                    break
                self.handle(record)
                if has_task_done:
                    q.task_done()
            except queue.Empty:
                break

    def stop(self):
        """
        Stop the listener.

        This asks the thread to terminate, and then waits for it to do so.
        Note that if you don't call this before your application exits, there
        may be some records still left on the queue, which won't be processed.
        """
        self._stop.set()
        self.queue.put_nowait(self._sentinel)
        self._thread.join()
        self._thread = None

class QueueLogger(object):
    def __init__(self):
        q = queue.Queue(-1)
        qh = QueueHandler(q)
        h = logging.StreamHandler()
        f = logging.Formatter('%(threadName)s: %(message)s')
        h.setFormatter(f)
        fl = logging.FileHandler("paegan.log")
        fl.setFormatter(f)
        self._listener = QueueListener(q, h, fl)
        self._logger = logging.getLogger('paegan')
        self._logger.setLevel(logging.INFO)
        self._logger.addHandler(qh)
        #self._logger.debug("hello")
        #log_listener.stop()

    def logger(self):
        return self._logger

    def listener(self):
        return self._listener

    def start(self):
        self._listener.start()

    def stop(self):
        self._listener.stop()

queue_logger = QueueLogger()

